from abc import ABCMeta, abstractmethod

from django.contrib.contenttypes.models import ContentType
from django.http import QueryDict
from django.http.response import JsonResponse
from django.utils.translation import gettext_lazy as _

from flag.exceptions import FlagBadRequest


class BaseMixin:
    __metaclass__ = ABCMeta
    api = False
    error = None

    def dispatch(self, request, *args, **kwargs):
        """
            Set `api=True` for rest framework.
            let rest framework handle the exception to choose the right renderer.
            validate method **should** be called in the derived API class.
        """
        if not self.api:
            try:
                self.validate(request)
            except FlagBadRequest as exc:
                return JsonResponse({'type': _('error'), 'detail': _(exc.detail)}, status=400)
        return super().dispatch(request, *args, **kwargs)

    @abstractmethod
    def validate(self, request, *args, **kwargs):
        pass

    def raise_error(self):
        raise FlagBadRequest(self.error)


class ContentTypeMixin(BaseMixin):
    model_obj = None
    data = None

    def _get_data_for_request(self, request):
        if self.api:
            data = request.POST or request.data
        else:
            data = request.POST

        if isinstance(data, QueryDict):
            return data.dict()
        return data

    def validate_data(self, request):
        data = self._get_data_for_request(request)
        if not data:
            self.error = 'no data passed'
            self.raise_error()
        return data

    def validate_app_name(self, app_name):
        if not app_name:
            self.error = 'app name is required'
            self.raise_error()

        if not ContentType.objects.filter(app_label=app_name).exists():
            self.error = f'{app_name} is not a valid app name'
            self.raise_error()

        return app_name

    def validate_model_name(self, model_name):
        if not model_name:
            self.error = 'model name is required'
            self.raise_error()

        return model_name

    def validate_model_id(self, model_id):
        if not model_id:
            self.error = 'model id is required'
            self.raise_error()
        try:
            model_id = int(model_id)
        except ValueError:
            self.error = f'model id must be an integer, "{model_id}" is NOT'
            self.raise_error()

        return model_id

    def validate_content_type_object(self, app_name, model_name):
        try:
            ct_object = ContentType.objects.get(model=model_name.lower(), app_label=app_name)
        except ContentType.DoesNotExist:
            self.error = f'"{model_name}" is NOT a valid model name'
            self.raise_error()

        return ct_object

    def validate_model_object(self, app_name, model_name, model_id):
        ct_object = self.validate_content_type_object(app_name, model_name)
        model_class = ct_object.model_class()
        model_query = model_class.objects.filter(id=model_id)
        if not model_query.exists() and model_query.count() != 1:
            self.error = f'"{model_id}" is NOT a valid model id for the model "{model_name}"'
            self.raise_error()
        return model_query.first()

    def validate(self, request, *args, **kwargs):
        super().validate(request, *args, **kwargs)
        data = self.validate_data(request)
        app_name = data.get('app_name', None)
        self.app_name = self.validate_app_name(app_name)

        model_name = data.get('model_name', None)
        self.model_name = self.validate_model_name(model_name)

        model_id = data.get('model_id', None)
        self.model_id = self.validate_model_id(model_id)
        self.model_obj = self.validate_model_object(self.app_name, self.model_name, self.model_id)

        # initialize data to be used by view
        self.data = data


class AJAXMixin(BaseMixin):
    def validate(self, request, *args, **kwargs):
        super().validate(request, *args, **kwargs)
        if not request.META.get('HTTP_X_REQUESTED_WITH', None) == 'XMLHttpRequest':
            self.error = 'Only AJAX requests are allowed'
            self.raise_error()



from django.contrib import admin

from flag.models import Flag, FlagInstance


class InlineFlagInstance(admin.TabularInline):
    model = FlagInstance
    extra = 0
    readonly_fields = ['flag', 'user', 'date_flagged', 'reason', 'info']


class FlaggedContentAdmin(admin.ModelAdmin):
    list_display = ['content_object', 'creator', 'state', 'moderator', 'count']
    readonly_fields = ['content_object', 'creator', 'count']
    exclude = ['content_type', 'object_id']
    search_fields = ['content_object']
    inlines = [InlineFlagInstance]


admin.site.register(Flag, FlaggedContentAdmin)



from flag.conf import settings
from tests.base import BaseFlagViewTest, Client, Flag, FlagInstance


class TestSetFlag(BaseFlagViewTest):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.post = cls.create_post()

    def setUp(self):
        super().setUp()
        self.client = Client(HTTP_X_REQUESTED_WITH='XMLHttpRequest')
        self.client.force_login(self.user_1)

    def test_flagging_successfully(self):
        data = self.data.copy()
        post = self.post
        data['model_id'] = post.id
        response = self.client.post(self.url, data=data)
        response_data = {
            'status': 0,
            'flag': 1,
            'msg': 'The content has been flagged successfully. A moderator will review it shortly.'
        }
        self.assertEqual(response.status_code, 200)
        self.assertDictEqual(response.json(), response_data)

        # check database
        flag = Flag.objects.get_flag(post)
        __, created = FlagInstance.objects.get_or_create(
                flag=flag,
                user=response.wsgi_request.user,
                reason=data['reason']
                )
        self.assertEqual(created, False)

    def test_flagging_flagged_object(self):
        data = self.data.copy()
        post = self.post
        data['model_id'] = post.id
        self.set_flag(post)
        response = self.client.post(self.url, data=data)
        response_data = {
            'status': 1,
            'msg': [f'This content has already been flagged by the user ({self.user_1.username})']
        }
        self.assertEqual(response.status_code, 200)
        self.assertDictEqual(response.json(), response_data)

    def test_flagging_unflagged_object(self):
        post = self.post
        data = self.data.copy()
        data['model_id'] = post.id
        data.pop('reason')
        response = self.client.post(self.url, data=data)
        response_data = {
            'status': 1,
            'msg': [f'This content has not been flagged by the user ({self.user_1.username})']
        }
        self.assertEqual(response.status_code, 200)
        self.assertDictEqual(response.json(), response_data)

    def test_unflagging_successfully(self):
        # un-flag => no reason is passed and the content must be already flagged by the user
        post = self.post
        self.set_flag(model_obj=post)
        data = self.data.copy()
        data['model_id'] = post.id
        data.pop('reason')
        response = self.client.post(self.url, data=data)
        response_data = {
            'status': 0,
            'msg': 'The content has been unflagged successfully.'
        }

        self.assertEqual(response.status_code, 200)
        self.assertDictEqual(response.json(), response_data)
        # check database
        flag = Flag.objects.get_flag(post)
        __, created = FlagInstance.objects.get_or_create(
                flag=flag,
                user=response.wsgi_request.user,
                reason=FlagInstance.reason_values[0]
                )
        self.assertEqual(created, True)

    def test_set_flag_for_unauthenticated_user(self):
        """Test whether unauthenticated user can create/delete flag using view"""
        self.client.logout()
        url = self.url
        response = self.client.post(url, data=self.data)

        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, '{}?next={}'.format(settings.LOGIN_URL, url))

    def test_incorrect_reason(self):
        """Test response when incorrect reason is passed"""
        data = self.data.copy()
        reason = -1
        data['reason'] = reason
        response = self.client.post(self.url, data=data)
        response_data = {
            'status': 1,
            'msg': [f'{reason} is an invalid reason']
        }

        self.assertEqual(response.status_code, 200)
        self.assertDictEqual(response.json(), response_data)

    def test_choosing_last_reason_without_info(self):
        """Test response when incorrect reason is passed"""
        data = self.data.copy()
        reason = FlagInstance.reason_values[-1]
        data.update({'reason': reason})
        response = self.client.post(self.url, data=data)
        response_data = {
            'status': 1,
            'msg': ['Please supply some information as the reason for flagging']
        }

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), response_data)

    def test_choosing_last_reason_with_info(self):
        """Test response when last reason is passed with info"""
        data = self.data.copy()
        post = self.post_2
        reason = FlagInstance.reason_values[-1]
        info = 'weird'
        data.update({'reason': reason, 'info': info, 'model_id': post.id})
        response = self.client.post(self.url, data=data)
        response_data = {
            'status': 0,
            'flag': 1,
            'msg': 'The content has been flagged successfully. A moderator will review it shortly.'
        }
        self.assertEqual(response.status_code, 200)
        self.assertDictEqual(response.json(), response_data)
        # check database
        flag = Flag.objects.get_flag(post)
        self.assertEqual(FlagInstance.objects.get(user=response.wsgi_request.user, flag=flag).info, info)



VERSION = (1, 0, 1)


def _get_version(version):
    if len(version) > 2:
        str_version = "%s.%s.%s" % version[:3]
    else:
        str_version = "%s.%s" % version[:2]

    return str_version


__version__ = _get_version(VERSION)


default_app_config = 'flag.apps.FlagConfig'



from unittest.mock import patch

from flag.signals import adjust_flagged_content
from flag.conf import settings
from tests.base import BaseFlagModelTest, Flag, FlagInstance


class FlagSignalTest(BaseFlagModelTest):
    def test_flagged_signal(self):
        user = self.user_2
        flag = self.create_flag(self.content_object_2, user)
        FlagInstance.objects.create_flag(user=user, flag=flag, reason=FlagInstance.reason_values[0], info='')
        flag.refresh_from_db()

        self.assertEqual(flag.count, 1)
        self.assertEqual(flag.state, Flag.State.FLAGGED.value)

        # instance edited won't increase the flag count
        flag_instance = FlagInstance.objects.get(user=user, flag=flag)
        self.assertIsNotNone(flag_instance)
        flag_instance.info = 'change value for test'
        flag_instance.save()
        flag.refresh_from_db()

        self.assertEqual(flag.count, 1)

    def test_unflagged_signal(self):
        user = self.user_1
        flag = self.create_flag(self.content_object_2, user)
        self.set_flag(model_obj=self.post_2, user=user)
        FlagInstance.objects.delete_flag(user=user, flag=flag)
        flag.refresh_from_db()

        self.assertEqual(flag.count, 0)
        self.assertEqual(flag.state, Flag.State.UNFLAGGED.value)

    @patch.object(settings, 'FLAG_ALLOWED', 1)
    def test_adjust_flagged_contents(self):
        post_1 = self.create_post()
        post_2 = self.create_post()
        flag_1 = Flag.objects.get_flag(post_1)
        flag_2 = Flag.objects.get_flag(post_2)

        self.assertFalse(flag_1.is_flagged)
        self.assertFalse(flag_2.is_flagged)

        self.set_flag(post_1, self.user_1)
        self.set_flag(post_1, self.user_2)

        self.set_flag(post_2, self.user_1)
        self.set_flag(post_2, self.user_2)

        flag_1.refresh_from_db()

        self.assertEqual(flag_1.count, 2)
        flag_2.refresh_from_db()

        self.assertEqual(flag_2.count, 2)

        # flagged post with wrong state will be adjusted
        flag_1.state = Flag.State.UNFLAGGED.value
        flag_1.save()
        flag_1.refresh_from_db()
        self.assertEqual(flag_1.state, Flag.State.UNFLAGGED.value)
        self.assertFalse(flag_1.is_flagged)

        # flagged post with right state => will be skipped
        flag_2.refresh_from_db()
        self.assertEqual(flag_2.state, Flag.State.FLAGGED.value)
        self.assertTrue(flag_2.is_flagged)

        adjust_flagged_content(self)
        flag_1.refresh_from_db()
        self.assertEqual(flag_1.count, 2)
        self.assertTrue(flag_1.is_flagged)



from django.views import View
from django.http.response import JsonResponse
from rest_framework.views import APIView

from tests.base import BaseFlagMixinsTest, Post
from flag.mixins import AJAXMixin, ContentTypeMixin


class MockedAJAXView(AJAXMixin, View):
    pass


class MockedContentTypeView(ContentTypeMixin, View):
    api = False

    def post(self, request, *args, **kwargs):
        return JsonResponse({})


class MockedContentTypeAPIView(ContentTypeMixin, APIView):
    api = True

    def post(self, request, *args, **kwargs):
        self.validate(request)
        return JsonResponse({})


class TestAJAXMixin(BaseFlagMixinsTest):
    def setUp(self):
        super().setUp()
        self.view = MockedAJAXView()

    def test_non_ajax_response_post_required(self):
        request = self.factory.post(self.url)
        response = self.view.dispatch(request)

        self.assertEqual(response.status_code, 400)
        self.assertEqual(self.view.error, 'Only AJAX requests are allowed')


class TestContentTypeMixin(BaseFlagMixinsTest):
    def setUp(self):
        super().setUp()
        self.view = MockedContentTypeView()

    def get_response(self, data):
        request = self.factory.post(self.url, data=data)
        return self.view.dispatch(request)

    def test_without_any_data(self):
        request = self.factory.post(self.url)
        response = self.view.dispatch(request)

        self.assertEqual(response.status_code, 400)
        self.assertEqual(self.view.error, 'no data passed')

    def test_without_app_name(self):
        data = self.data.copy()
        data.pop('app_name')
        response = self.get_response(data)

        self.assertEqual(response.status_code, 400)
        self.assertEqual(self.view.error, 'app name is required')

    def test_without_model_name(self):
        data = self.data.copy()
        data.pop('model_name')
        response = self.get_response(data)

        self.assertEqual(response.status_code, 400)
        self.assertEqual(self.view.error, 'model name is required')

    def test_without_model_id(self):
        data = self.data.copy()
        data.pop('model_id')
        response = self.get_response(data)

        self.assertEqual(response.status_code, 400)
        self.assertEqual(self.view.error, 'model id is required')

    def test_invalid_app_name(self):
        data = self.data.copy()
        val = 'not exists'
        data['app_name'] = val
        response = self.get_response(data)

        self.assertEqual(response.status_code, 400)
        self.assertEqual(self.view.error, f'{val} is not a valid app name')

    def test_invalid_model_name(self):
        data = self.data.copy()
        val = 'not exists'
        data['model_name'] = val
        response = self.get_response(data)

        self.assertEqual(response.status_code, 400)
        self.assertEqual(self.view.error, f'"{val}" is NOT a valid model name')

    def test_model_id_which_doesnt_exist(self):
        data = self.data.copy()
        val = 100
        data['model_id'] = val
        response = self.get_response(data)

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            self.view.error,
            f'"{val}" is NOT a valid model id for the model "{data["model_name"]}"')

    def test_non_integral_model_id(self):
        data = self.data.copy()
        val = 'c'
        data['model_id'] = val
        response = self.get_response(data)

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            self.view.error,
            f'model id must be an integer, "{val}" is NOT')

    def test_success(self):
        data = self.data.copy()
        data['model_id'] = str(data['model_id'])
        response = self.get_response(data)

        self.assertEqual(response.status_code, 200)
        self.assertIsNone(self.view.error)
        self.assertEqual(self.view.data, data)
        self.assertEqual(self.view.model_name, data['model_name'])
        self.assertEqual(self.view.app_name, data['app_name'])
        self.assertEqual(self.view.model_id, int(data['model_id']))
        self.assertEqual(self.view.model_obj, Post.objects.get(id=data['model_id']))


class TestContentTypeMixinForAPI(BaseFlagMixinsTest):
    def setUp(self):
        super().setUp()
        self.view = MockedContentTypeAPIView()

    def get_response(self, data):
        request = self.factory.post(self.url, data=data)
        return self.view.dispatch(request)

    def test_without_any_data(self):
        request = self.factory.post(self.url)
        response = self.view.dispatch(request)

        self.assertEqual(response.status_code, 400)
        self.assertEqual(self.view.error, 'no data passed')

    def test_without_app_name(self):
        data = self.data.copy()
        data.pop('app_name')
        response = self.get_response(data)

        self.assertEqual(response.status_code, 400)
        self.assertEqual(self.view.error, 'app name is required')

    def test_without_model_name(self):
        data = self.data.copy()
        data.pop('model_name')
        response = self.get_response(data)

        self.assertEqual(response.status_code, 400)
        self.assertEqual(self.view.error, 'model name is required')

    def test_without_model_id(self):
        data = self.data.copy()
        data.pop('model_id')
        response = self.get_response(data)

        self.assertEqual(response.status_code, 400)
        self.assertEqual(self.view.error, 'model id is required')

    def test_invalid_app_name(self):
        data = self.data.copy()
        val = 'not exists'
        data['app_name'] = val
        response = self.get_response(data)

        self.assertEqual(response.status_code, 400)
        self.assertEqual(self.view.error, f'{val} is not a valid app name')

    def test_invalid_model_name(self):
        data = self.data.copy()
        val = 'not exists'
        data['model_name'] = val
        response = self.get_response(data)

        self.assertEqual(response.status_code, 400)
        self.assertEqual(self.view.error, f'"{val}" is NOT a valid model name')

    def test_model_id_which_doesnt_exist(self):
        data = self.data.copy()
        val = 100
        data['model_id'] = val
        response = self.get_response(data)

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            self.view.error,
            f'"{val}" is NOT a valid model id for the model "{data["model_name"]}"')

    def test_non_integral_model_id(self):
        data = self.data.copy()
        val = 'c'
        data['model_id'] = val
        response = self.get_response(data)

        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            self.view.error,
            f'model id must be an integer, "{val}" is NOT')

    def test_success(self):
        data = self.data.copy()
        data['model_id'] = str(data['model_id'])
        response = self.get_response(data)

        self.assertEqual(response.status_code, 200)
        self.assertIsNone(self.view.error)
        self.assertEqual(self.view.data, data)
        self.assertEqual(self.view.model_name, data['model_name'])
        self.assertEqual(self.view.app_name, data['app_name'])
        self.assertEqual(self.view.model_id, int(data['model_id']))
        self.assertEqual(self.view.model_obj, Post.objects.get(id=data['model_id']))



from django.contrib.auth.mixins import LoginRequiredMixin
from django.http.response import JsonResponse
from django.views import View

from flag.mixins import AJAXMixin, ContentTypeMixin
from flag.utils import process_flagging_request


class SetFlag(LoginRequiredMixin, ContentTypeMixin, AJAXMixin, View):
    def post(self, request, *args, **kwargs):
        response = process_flagging_request(user=request.user, model_obj=self.model_obj, data=self.data)
        return JsonResponse(response)



from unittest.mock import patch

from django.contrib.contenttypes.models import ContentType

from flag.utils import get_content_type, get_model_object, process_flagging_request, get_user_for_model
from tests.base import BaseFlagTest, Post


class TestGetContentType(BaseFlagTest):
    def test_success(self):
        post = self.post_1
        response = get_content_type(post)
        self.assertEqual(response, ContentType.objects.get_for_model(post.__class__))


class TestGetModelObject(BaseFlagTest):
    def test_success(self):
        data = self.data.copy()
        data.pop('info')
        data.pop('reason')
        response = get_model_object(**data)
        post = Post.objects.get(id=data['model_id'])
        self.assertEqual(response, post)


class TestGetUserForModel(BaseFlagTest):
    def test_when_user_field_is_associated(self):
        response = get_user_for_model(self.post_1)
        self.assertEqual(response, self.post_1.user)

    @patch('testapp.post.models.Post')
    def test_when_user_field_is_not_associated(self, mocked_post_model):
        mocked_post = mocked_post_model()
        response = get_user_for_model(mocked_post)
        self.assertEqual(response, None)


class TestProcessFlaggingRequest(BaseFlagTest):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.post = cls.create_post()

    def test_flagging_successfully(self):
        data = self.data.copy()
        post = self.post
        data['model_id'] = post.id
        response = process_flagging_request(user=self.user_1, model_obj=post, data=data)
        response_data = {
            'status': 0,
            'flag': 1,
            'msg': 'The content has been flagged successfully. A moderator will review it shortly.'
        }
        self.assertDictEqual(response, response_data)

    def test_unflagging_successfully(self):
        post = self.post
        self.set_flag(model_obj=post)
        data = self.data.copy()
        data['model_id'] = post.id
        data.pop('reason')
        response = process_flagging_request(user=self.user_1, model_obj=post, data=data)
        response_data = {
            'status': 0,
            'msg': 'The content has been unflagged successfully.'
        }

        self.assertDictEqual(response, response_data)

    def test_flagging_flagged_object(self):
        data = self.data.copy()
        post = self.post
        data['model_id'] = post.id
        self.set_flag(post)
        response = process_flagging_request(user=self.user_1, model_obj=self.post, data=data)
        response_data = {
            'status': 1,
            'msg': [f'This content has already been flagged by the user ({self.user_1.username})']
        }
        self.assertDictEqual(response, response_data)

    def test_flagging_unflagged_object(self):
        post = self.post
        data = self.data.copy()
        data['model_id'] = post.id
        data.pop('reason')
        response = process_flagging_request(user=self.user_1, model_obj=self.post, data=data)
        response_data = {
            'status': 1,
            'msg': [f'This content has not been flagged by the user ({self.user_1.username})']
        }
        self.assertDictEqual(response, response_data)



from django.utils.translation import gettext_lazy as _

# reason displayed when flagging an object
FLAG_REASONS = [
        (1, _("Spam | Exists only to promote a service ")),
        (2, _("Abusive | Intended at promoting hatred")),
    ]

# number of flags before an object is marked as flagged
FLAG_ALLOWED = 10



from django.contrib.auth import get_user_model
from rest_framework import serializers

from flag.models import Flag


User = get_user_model()


class UserSerializer(serializers.ModelSerializer):
    class Meta:
        model = User
        fields = ('id', 'username')
        lookup_field = 'username'


class FlagSerializer(serializers.ModelSerializer):
    class Meta:
        model = Flag
        fields = ('creator', 'count', 'state', 'verbose_state', 'moderator', 'is_flagged', 'reporters')
        read_only_fields = ('creator', 'moderator', 'reporters')

    creator = UserSerializer()
    moderator = UserSerializer()
    verbose_state = serializers.SerializerMethodField()
    reporters = serializers.SerializerMethodField()

    def __init__(self, *args, **kwargs):
        context = kwargs['context']
        self.model_obj = context.get('model_obj')
        self.user = context.get('user')
        super().__init__(*args, **kwargs)

    @staticmethod
    def get_verbose_state(obj):
        return obj.get_verbose_state(obj.state)

    def get_reporters(self, obj):
        flag_obj = Flag.objects.get_flag(self.model_obj)
        return [
            {'id': instance.user.id, 'username': instance.user.username} for instance in flag_obj.flags.all()
        ]



from rest_framework import status

from flag.api.serializers import FlagSerializer
from tests.base import BaseFlagAPITest, Flag, FlagInstance


class FlagAPIViewsTest(BaseFlagAPITest):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.post = cls.create_post()
        cls.flag = Flag.objects.get_flag(cls.post)

    def setUp(self):
        super().setUp()
        self.url = '/api/flag/'
        self.init_count = 0
        self.context = {
            'model_obj': self.post,
            'user': self.user_1
        }

    def test_flagging_successfully_with_url_encoded_form(self):
        data = self.data.copy()
        post = self.post
        data['model_id'] = post.id
        self.flag.refresh_from_db()
        init_count = self.flag.count

        response = self.client.post(self.url, data=data)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.flag.refresh_from_db()
        self.assertEqual(response.data, FlagSerializer(self.flag, context=self.context).data)

        # check database
        __, created = FlagInstance.objects.get_or_create(
                flag=self.flag,
                user=response.wsgi_request.user,
                reason=data['reason']
                )
        self.assertEqual(created, False)
        self.assertEqual(self.flag.count, init_count + 1)

    def test_flagging_successfully_with_json_format(self):
        data = self.data.copy()
        post = self.post
        data['model_id'] = post.id
        response = self.client.post(self.url, data=data, format='json')

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.flag.refresh_from_db()
        self.assertEqual(response.data, FlagSerializer(self.flag, context=self.context).data)

        # check database
        __, created = FlagInstance.objects.get_or_create(
                flag=self.flag,
                user=response.wsgi_request.user,
                reason=data['reason']
                )
        self.assertEqual(created, False)
        self.assertEqual(self.flag.count, self.init_count + 1)

    def test_flagging_flagged_object(self):
        data = self.data.copy()
        post = self.post
        data['model_id'] = post.id
        self.set_flag(post)
        response = self.client.post(self.url, data=data)

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            response.json()['detail'],
            [f'This content has already been flagged by the user ({self.user_1.username})']
        )

    def test_flagging_unflagged_object(self):
        post = self.post
        data = self.data.copy()
        data['model_id'] = post.id
        data.pop('reason')
        response = self.client.post(self.url, data=data)

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            response.json()['detail'],
            [f'This content has not been flagged by the user ({self.user_1.username})']
            )

    def test_unflagging_successfully(self):
        # un-flag => no reason is passed and the content must be already flagged by the user
        post = self.post
        self.flag.refresh_from_db()
        init_count = self.flag.count
        self.set_flag(post)
        self.flag.refresh_from_db()
        self.assertEqual(self.flag.count, init_count + 1)

        data = self.data.copy()
        data['model_id'] = post.id
        data.pop('reason')
        response = self.client.post(self.url, data=data)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.flag.refresh_from_db()
        self.assertEqual(response.data, FlagSerializer(self.flag, context=self.context).data)
        # check database
        self.assertEqual(self.flag.count, init_count)
        __, created = FlagInstance.objects.get_or_create(
                flag=self.flag,
                user=response.wsgi_request.user,
                reason=FlagInstance.reason_values[0]
                )
        self.assertEqual(created, True)

    def test_set_flag_for_unauthenticated_user(self):
        """Test whether unauthenticated user can create/delete flag using view"""
        self.client.logout()
        url = self.url
        response = self.client.post(url, data=self.data)

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_incorrect_reason(self):
        """Test response when incorrect reason is passed"""
        data = self.data.copy()
        reason = -1
        data['reason'] = reason
        response = self.client.post(self.url, data=data)

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json()['detail'], [f'{reason} is an invalid reason'])

    def test_choosing_last_reason_without_info(self):
        """Test response when incorrect reason is passed"""
        data = self.data.copy()
        reason = FlagInstance.reason_values[-1]
        data.update({'reason': reason})
        response = self.client.post(self.url, data=data)

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json()['detail'], ['Please supply some information as the reason for flagging'])

    def test_choosing_last_reason_with_info(self):
        """Test response when last reason is passed with info"""
        data = self.data.copy()
        post = self.post_2
        flag = Flag.objects.get_flag(post)
        init_count = flag.count
        reason = FlagInstance.reason_values[-1]
        info = 'weird'
        data.update({'reason': reason, 'info': info, 'model_id': post.id})
        response = self.client.post(self.url, data=data)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        flag.refresh_from_db()
        context = self.context.copy()
        context['model_obj'] = post
        self.assertEqual(response.data, FlagSerializer(flag, context=context).data)
        # check database
        self.assertEqual(FlagInstance.objects.get(user=response.wsgi_request.user, flag=flag).info, info)
        self.assertEqual(flag.count, init_count + 1)



# Generated by Django 3.1.1 on 2020-09-15 19:41

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='Flag',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('object_id', models.PositiveIntegerField()),
                ('state', models.SmallIntegerField(choices=[(1, 'Unflagged'), (2, 'Flagged'), (3, 'Flag rejected by the moderator'), (5, 'Content modified by the author'), (4, 'Creator notified')], default=1)),
                ('count', models.PositiveIntegerField(default=0)),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype')),
                ('creator', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='flags_against', to=settings.AUTH_USER_MODEL)),
                ('moderator', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='flags_moderated', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'Flag',
                'unique_together': {('content_type', 'object_id')},
            },
        ),
        migrations.CreateModel(
            name='FlagInstance',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date_flagged', models.DateTimeField(auto_now_add=True)),
                ('reason', models.SmallIntegerField(choices=[(1, 'Spam | Exists only to promote a service '), (2, 'Abusive | Intended at promoting hatred'), (100, 'Something else')], default=1)),
                ('info', models.TextField(blank=True, null=True)),
                ('flag', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='flags', to='flag.flag')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='flags_by', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'Flag Instance',
                'verbose_name_plural': 'Flag Instances',
                'ordering': ['-date_flagged'],
                'unique_together': {('flag', 'user')},
            },
        ),
    ]



from collections import namedtuple
from enum import IntEnum, unique

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from flag.managers import FlagInstanceManager, FlagManager
from flag.conf import settings

User = settings.AUTH_USER_MODEL


class Flag(models.Model):
    """Used to add flag/moderation to a model"""
    @unique
    class State(IntEnum):
        UNFLAGGED = 1
        FLAGGED = 2
        REJECTED = 3
        NOTIFIED = 4
        RESOLVED = 5

    STATE_CHOICES = [
        (State.UNFLAGGED.value, _('Unflagged')),
        (State.FLAGGED.value, _('Flagged')),
        (State.REJECTED.value, _('Flag rejected by the moderator')),
        (State.RESOLVED.value, _('Content modified by the author')),
        (State.NOTIFIED.value, _('Creator notified')),
    ]

    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    object_id = models.PositiveIntegerField()
    content_object = GenericForeignKey('content_type', 'object_id')
    creator = models.ForeignKey(User, related_name='flags_against', null=True, on_delete=models.CASCADE)
    state = models.SmallIntegerField(choices=STATE_CHOICES, default=State.UNFLAGGED.value)
    moderator = models.ForeignKey(User, null=True, related_name='flags_moderated', on_delete=models.SET_NULL)
    count = models.PositiveIntegerField(default=0)

    objects = FlagManager()

    class Meta:
        verbose_name = _('Flag')
        unique_together = ['content_type', 'object_id']

    def increase_count(self):
        field = 'count'
        self.refresh_from_db()
        self.count = models.F(field) + 1
        self.save(update_fields=[field])

    def decrease_count(self):
        field = 'count'
        self.refresh_from_db()
        self.count = models.F(field) - 1
        self.save(update_fields=[field])

    def get_clean_state(self, state):
        err = ValidationError(_('%(state)s is an invalid state'), code='invalid', params={'state': state})
        try:
            state = int(state)
            if state not in [st.value for st in self.State]:
                raise err
        except (ValueError, TypeError):
            raise err
        return state

    def get_verbose_state(self, state):
        state = self.get_clean_state(state)
        for item in self.STATE_CHOICES:
            if item[0] == state:
                return item[1]

    def toggle_state(self, state, moderator):
        state = self.get_clean_state(state)
        # toggle states occurs between rejected and resolved states only
        if state != self.State.REJECTED.value and state != self.State.RESOLVED.value:
            raise ValidationError(_('%(state)s is an invalid state'), code='invalid', params={'state': state})
        if self.state == state:
            self.state = self.State.FLAGGED.value
        else:
            self.state = state
        self.moderator = moderator
        self.save()

    def toggle_flagged_state(self):
        allowed_flags = settings.FLAG_ALLOWED
        self.refresh_from_db()
        field = 'state'
        if self.count > allowed_flags and (
            getattr(self, field) not in [self.State.RESOLVED.value, self.State.REJECTED.value]
        ):
            setattr(self, field, self.State.FLAGGED.value)
        else:
            setattr(self, field, self.State.UNFLAGGED.value)
        self.save(update_fields=[field])

    @property
    def is_flagged(self):
        return self.state != self.State.UNFLAGGED.value


class FlagInstance(models.Model):
    REASON = settings.FLAG_REASONS

    REASON.append((100, _('Something else')))

    # Make a named tuple
    Reasons = namedtuple('Reason', ['value', 'reason'])

    # Construct the list of named tuples
    reasons = []
    for reason in REASON:
        reasons.append(Reasons(*reason))

    reason_values = [reason.value for reason in reasons]

    flag = models.ForeignKey(Flag, related_name='flags', on_delete=models.CASCADE)
    user = models.ForeignKey(User, related_name='flags_by', on_delete=models.CASCADE)
    date_flagged = models.DateTimeField(auto_now_add=timezone.now)
    reason = models.SmallIntegerField(choices=REASON, default=reason_values[0])
    info = models.TextField(null=True, blank=True)

    objects = FlagInstanceManager()

    class Meta:
        verbose_name = _('Flag Instance')
        verbose_name_plural = _('Flag Instances')
        unique_together = ['flag', 'user']
        ordering = ['-date_flagged']

    def clean(self):
        """If something else is choosen, info shall not be empty"""
        if self.reason == self.reason_values[-1] and not self.info:
            raise ValidationError(
                {
                    'info': ValidationError(
                        _('Please provide some information why you choose to report the content'),
                        code='required')
                }
            )

    def save(self, *args, **kwargs):
        self.clean()
        super(FlagInstance, self).save(*args, **kwargs)



from django.apps import apps
from django.core.exceptions import ValidationError
from django.db import IntegrityError, models
from django.utils.translation import gettext_lazy as _

from flag.utils import get_content_type, get_user_for_model


class FlagManager(models.Manager):
    def get_flag(self, model_obj):
        ctype = get_content_type(model_obj)
        creator = get_user_for_model(model_obj)
        flag, __ = self.get_or_create(content_type=ctype, object_id=model_obj.id, creator=creator)
        return flag

    def is_flagged(self, model_obj):
        flag = self.get_flag(model_obj)
        return flag.is_flagged

    def has_flagged(self, user, model_obj):
        """
        Returns whether a model object has been flagged by a user or not

        Args:
            user (object): the user to be inquired about.
            model_obj (object): the model object to be inquired upon.

        Returns:
            bool
        """
        flag = self.get_flag(model_obj)
        return flag.flags.filter(user=user).exists()


class FlagInstanceManager(models.Manager):
    def _clean_reason(self, reason):
        err = ValidationError(
                _('%(reason)s is an invalid reason'),
                params={'reason': reason},
                code='invalid'
                )
        try:
            reason = int(reason)
            if reason in self.model.reason_values:
                return reason
            raise err

        except (ValueError, TypeError):
            raise err

    def _clean(self, reason, info):
        cleaned_reason = self._clean_reason(reason)
        cleaned_info = None

        if cleaned_reason == self.model.reason_values[-1]:
            cleaned_info = info
            if not cleaned_info:
                raise ValidationError(
                    _('Please supply some information as the reason for flagging'),
                    params={'info': info},
                    code='required'
                )
        return cleaned_reason, cleaned_info

    def create_flag(self, user, flag, reason, info):
        cleaned_reason, cleaned_info = self._clean(reason, info)
        try:
            self.create(flag=flag, user=user, reason=cleaned_reason, info=cleaned_info)
        except IntegrityError:
            raise ValidationError(
                    _('This content has already been flagged by the user (%(user)s)'),
                    params={'user': user},
                    code='invalid'
                )

    def delete_flag(self, user, flag):
        try:
            self.get(user=user, flag=flag).delete()
        except self.model.DoesNotExist:
            raise ValidationError(
                _('This content has not been flagged by the user (%(user)s)'),
                params={'user': user},
                code='invalid'
            )

    def set_flag(self, user, model_obj, **kwargs):
        Flag = apps.get_model('flag', 'Flag')
        flag_obj = Flag.objects.get_flag(model_obj)
        info = kwargs.get('info', None)
        reason = kwargs.get('reason', None)

        if reason:
            self.create_flag(user, flag_obj, reason, info)
            created = True
        else:
            self.delete_flag(user, flag_obj)
            created = False

        return created



from django.urls import path
from rest_framework.urlpatterns import format_suffix_patterns

from flag.api import views


urlpatterns = [
    path('flag/', views.SetFlag.as_view(),  name='flag'),
]

urlpatterns = format_suffix_patterns(urlpatterns)



"""General purpose functions that provide utility throughout the application"""
from django.contrib.auth import get_user_model
from django.apps import apps
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _


def get_content_type(model_obj):
    return ContentType.objects.get_for_model(model_obj.__class__)


def get_model_object(*, app_name, model_name, model_id):
    """
    Get content object.
    Args:
        app_name (str): name of the app that contains the model.
        model_name (str): name of the model class.
        model_id (int): the id of the model object.

    Returns:
        object: model object according to the parameters passed
    """
    content_type = ContentType.objects.get(app_label=app_name, model=model_name.lower())
    model_object = content_type.get_object_for_this_type(id=model_id)

    return model_object


def get_user_for_model(obj):
    User = get_user_model()
    for field in obj._meta.fields:
        if field.related_model == User:
            return getattr(obj, field.name)


def process_flagging_request(*, user, model_obj, data):
    """
    Process flagging request and return a response. This handles request for both Django and DRF

    Args:
        user ([type]): The looged in user
        model_obj ([type]): the object being flagged
        data (dict): the data received from the request

    Returns:
        [dict]: response has three keys:
            `status`(int): Non-zero indicates the request failed due to `ValidationError`.
            `msg`(str): response, success message in case request succeeds, reason for
            failure if it doesn't.

            **This key will only be present when request succeeds.**
            `flag`(int): Non-Zero(1) indicates that flag is created.
    """
    FlagInstance = apps.get_model('flag', 'FlagInstance')

    response = {'status': 1}
    try:
        if FlagInstance.objects.set_flag(user, model_obj, **data):
            response['msg'] = _(
                'The content has been flagged successfully. '
                'A moderator will review it shortly.'
                )
            response['flag'] = 1
        else:
            response['msg'] = _('The content has been unflagged successfully.')

        response.update({
            'status': 0
        })
    except ValidationError as e:
        response.update({
            'msg': e.messages
        })

    return response



from django import template
from django.core.exceptions import ImproperlyConfigured
from django.utils.translation import gettext_lazy as _

from flag.models import Flag, FlagInstance
from flag.conf import settings

register = template.Library()


def get_app_name(obj):
    return type(obj)._meta.app_label


def get_model_name(obj):
    return type(obj).__name__


def has_flagged(user, obj):
    if user.is_authenticated:
        return Flag.objects.has_flagged(user, obj)

    return False


@register.simple_tag(name='get_login_url')
def get_login_url():
    login_url = getattr(settings, 'LOGIN_URL', None)
    if not login_url:
        raise ImproperlyConfigured(_('Django Flag App: LOGIN_URL is not defined in the settings'))
    if not login_url.endswith('/'):
        login_url += '/'
    return login_url


@register.inclusion_tag('flag/flag_form.html')
def render_flag_form(obj, user):
    """
    A template tag used for adding flag form in templates

    To render the flag form for a post model inside the app posts

    Usage: `{% render_flag_form post user %}`
    """
    return {
        'app_name': get_app_name(obj),
        'model_name': get_model_name(obj),
        'model_id': obj.id,
        'user': user,
        'has_flagged': has_flagged(user, obj),
        'flag_reasons': FlagInstance.reasons
    }



from unittest.mock import patch

from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.contrib.contenttypes.models import ContentType
from django.shortcuts import reverse
from django.test import Client, RequestFactory, TestCase
from rest_framework.test import APITestCase

from flag.models import Flag, FlagInstance
from testapp.post.models import Post


User = get_user_model()


class BaseFlagTestUtils:
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.user_1 = User.objects.create_user(
                    username='test-1',
                    email='a@a.com',
                    password='1234'
        )
        cls.user_2 = User.objects.create_user(
            username='test-2',
            email='b@b.com',
            password='1234'
        )
        cls.moderator = User.objects.create_user(
            username='moderator',
            email='b@b.com',
            password='1234'
        )
        moderator_group = Group.objects.filter(name='flag_moderator').first()
        moderator_group.user_set.add(cls.moderator)
        cls.posts = 0
        cls.post_1 = cls.create_post()
        cls.post_2 = cls.create_post()
        content_type = ContentType.objects.get(model=type(cls.post_1).__name__.lower())
        cls.content_object_1 = content_type.get_object_for_this_type(id=cls.post_1.id)
        cls.content_object_2 = content_type.get_object_for_this_type(id=cls.post_2.id)
        cls.flags = 0

    def setUp(self):
        super().setUp()
        self.client.force_login(self.user_1)
        self.url = reverse('flag:flag')
        self.data = {
            'app_name': 'post',
            'model_name': 'Post',
            'model_id': self.post_1.id,
            'reason': FlagInstance.reason_values[0],
            'info': ''
        }
        self.addCleanup(patch.stopall)

    @classmethod
    def create_post(cls):
        cls.posts += 1
        return Post.objects.create(
            user=cls.user_1,
            title=f'post {cls.posts}',
            body=f'post number {cls.posts} body'
        )

    @classmethod
    def increase_flag_count(cls):
        cls.flags = 1

    @classmethod
    def create_flag(cls, ct_object=None, creator=None):
        if not ct_object:
            ct_object = cls.content_object_1
        if not creator:
            creator = cls.user_1
        return Flag.objects.create(content_object=ct_object, creator=creator)

    @classmethod
    def set_flag(cls, model_obj=None, user=None, reason=None, info=None):
        if not user:
            user = cls.user_1
        if not reason:
            reason = FlagInstance.reason_values[0]
        if not info:
            info = None
        if not model_obj:
            model_obj = cls.post_1
        flag_obj = Flag.objects.get_flag(model_obj)
        cls.increase_flag_count()
        return FlagInstance.objects.create(
            flag=flag_obj,
            user=user,
            reason=reason,
            info=info
        )


class BaseFlagTest(BaseFlagTestUtils, TestCase):
    pass


class BaseFlagModelTest(BaseFlagTest):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.flag = cls.create_flag()


class BaseFlagViewTest(BaseFlagTest):
    def setUp(self):
        super().setUp()
        self.client = Client(HTTP_X_REQUESTED_WITH='XMLHttpRequest')
        self.client.force_login(self.user_1)

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.flag_instance = cls.set_flag()


class BaseTemplateTagsTest(BaseFlagTest):
    class MockUser:
        """Mock unauthenticated user for template. The User instance always returns True for `is_authenticated`"""
        is_authenticated = False

    def setUp(self):
        super().setUp()
        self.factory = RequestFactory()


class BaseFlagMixinsTest(BaseFlagTest):
    def setUp(self):
        super().setUp()
        self.data = {
            'app_name': 'post',
            'model_name': 'Post',
            'model_id': self.post_1.id
        }
        self.factory = RequestFactory()


class BaseFlagAPITest(BaseFlagTestUtils, APITestCase):
    pass



from django.contrib.auth.models import Permission, Group
from django.contrib.contenttypes.models import ContentType
from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver
from django.utils.translation import gettext_lazy as _

from flag.models import Flag, FlagInstance


@receiver(post_save, sender=FlagInstance)
def flagged(sender, instance, created, raw, using, update_fields, **kwargs):
    """Increase flag count in the flag model after creating an instance"""
    if created:
        instance.flag.increase_count()
        instance.flag.toggle_flagged_state()


@receiver(post_delete, sender=FlagInstance)
def unflagged(sender, instance, using, **kwargs):
    """Decrease flag count in the flag model before deleting an instance"""
    instance.flag.decrease_count()
    instance.flag.toggle_flagged_state()


def create_permission_groups(sender, **kwargs):
    flag_ct = ContentType.objects.get_for_model(Flag)
    delete_flagged_perm, __ = Permission.objects.get_or_create(
        codename='delete_flagged_content',
        name=_('Can delete flagged content'),
        content_type=flag_ct
    )
    moderator_group, __ = Group.objects.get_or_create(name='flag_moderator')
    moderator_group.permissions.add(delete_flagged_perm)


def adjust_flagged_content(sender, **kwargs):
    for flag in Flag.objects.all():
        flag.toggle_flagged_state()



from rest_framework.views import APIView
from rest_framework import permissions, status
from rest_framework.response import Response

from flag.mixins import ContentTypeMixin
from flag.api.serializers import FlagSerializer
from flag.models import Flag
from flag.utils import process_flagging_request


class SetFlag(ContentTypeMixin, APIView):
    api = True
    permission_classes = (permissions.IsAuthenticated,)

    def _get_serializer_context(self):
        context = {}
        context['model_obj'] = self.model_obj
        context['user'] = self.request.user
        return context

    def post(self, request, *args, **kwargs):
        self.validate(self.request)
        response = process_flagging_request(user=request.user, model_obj=self.model_obj, data=self.data)
        detail = {'detail': response['msg']}

        if response['status']:  # 1 indicates bad request
            return Response(status=status.HTTP_400_BAD_REQUEST, data=detail)

        flag = Flag.objects.get_flag(self.model_obj)
        serializer = FlagSerializer(instance=flag, context=self._get_serializer_context())
        return Response(serializer.data, status=status.HTTP_200_OK)



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _
from django.db.models.signals import post_migrate


class FlagConfig(AppConfig):
    name = 'flag'
    verbose_name = _('flag')

    def ready(self):
        import flag.signals

        post_migrate.connect(flag.signals.create_permission_groups, sender=self)
        post_migrate.connect(flag.signals.adjust_flagged_content, sender=self)



import codecs
import os
import setuptools
import importlib

BASE_DIR = os.path.dirname(os.path.abspath(__file__))


def get_version(rel_path):
    return importlib.import_module('flag').__version__


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    # intentionally *not* adding an encoding option to open, See:
    #   https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


setuptools.setup(
    name='django-flag-app',
    version=get_version('flag/__init__.py'),
    author='Abhyudai',
    author_email='',
    description='A pluggable django application that adds the ability for users to flag(or report) your models',
    long_description=read('README.rst'),
    url='https://github.com/abhiabhi94/django-flag-app',
    project_urls={
        'Documentation': 'https://django-flag-app.readthedocs.io',
        'Source Code': 'https://github.com/abhiabhi94/django-flag-app',
    },
    packages=setuptools.find_packages(exclude=['docs', 'test*']),
    include_package_data=True,
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 2.1',
        'Framework :: Django :: 2.2',
        'Framework :: Django :: 3.0',
        'Framework :: Django :: 3.1',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
    python_requires='>=3.6',
    install_requires=['django'],
    keywords='django flag report',
    zip_safe=False,
)



import pytest

from flag import _get_version


class TestVersion:
    @pytest.mark.parametrize(('version', 'result'), (
        ((1, 2, 3), ('1.2.3')),  # with patch
        ((1, 2), ('1.2')),  # without patch
    ))
    def test_get_version(self, version, result):
        assert _get_version((version)) == result



from unittest.mock import patch

from django.core.exceptions import ImproperlyConfigured

from flag.conf import settings
from flag.templatetags.flag_tags import get_app_name, get_model_name, has_flagged, render_flag_form, get_login_url
from tests.base import BaseTemplateTagsTest, FlagInstance


class TestFlagTemplateTest(BaseTemplateTagsTest):
    def test_get_model_name(self):
        self.assertEqual(get_model_name(self.post_1), 'Post')

    def test_get_app_name(self):
        self.assertEqual(get_app_name(self.post_1), 'post')

    def test_has_flagged_for_unauthenticated_user(self):
        user = self.MockUser()
        self.assertEqual(has_flagged(user, self.post_1), False)

    def test_has_flagged_for_authenticated_user(self):
        user = self.user_2
        post = self.post_1
        self.assertEqual(has_flagged(user, post), False)
        # flag the object
        self.set_flag(post, user)
        self.assertEqual(has_flagged(user, post), True)

    @patch.object(settings, 'LOGIN_URL', None)
    def test_login_url_without_setting_login_url(self):
        with self.assertRaises(ImproperlyConfigured) as error:
            get_login_url()
        self.assertIsInstance(error.exception, ImproperlyConfigured)

    @patch.object(settings, 'LOGIN_URL', '/login')
    def test_get_login_url_without_backward_slash(self):
        self.assertEqual(settings.LOGIN_URL + '/', get_login_url())

    @patch.object(settings, 'LOGIN_URL', '/login/')
    def test_login_url_with_backward_slash(self):
        self.assertEqual(get_login_url(), settings.LOGIN_URL)

    def test_render_flag_form(self):
        post = self.post_1
        user = self.user_2
        data = render_flag_form(post, user)

        self.assertEqual(data['app_name'], post._meta.app_label)
        self.assertEqual(data['model_name'], type(post).__name__)
        self.assertEqual(data['model_id'], post.id)
        self.assertEqual(data['user'], user)
        self.assertEqual(data['flag_reasons'], FlagInstance.reasons)
        self.assertEqual(data['has_flagged'], False)

        # flag the object
        self.set_flag(post, user)
        data = render_flag_form(post, user)

        self.assertEqual(data['app_name'], post._meta.app_label)
        self.assertEqual(data['model_name'], type(post).__name__)
        self.assertEqual(data['model_id'], post.id)
        self.assertEqual(data['user'], user)
        self.assertEqual(data['flag_reasons'], FlagInstance.reasons)
        self.assertEqual(data['has_flagged'], True)



from unittest.mock import patch

from django.core.exceptions import ValidationError

from flag.conf import settings
from tests.base import BaseFlagModelTest, Flag, FlagInstance


class FlagModelTest(BaseFlagModelTest):
    def test_can_create_flag(self):
        self.assertIsNotNone(Flag.objects.create(content_object=self.post_2))

    def test_increase_count(self):
        self.flag.increase_count()
        self.flag.refresh_from_db()

        self.assertEqual(self.flag.count, 1)

    def test_decrease_count(self):
        self.flag.increase_count()
        self.flag.decrease_count()
        self.flag.refresh_from_db()

        self.assertEqual(self.flag.count, 0)

    def test_get_clean_state(self):
        flag = self.flag
        state = flag.get_clean_state(Flag.State.FLAGGED.value)
        self.assertEqual(state, Flag.State.FLAGGED.value)

        # int not in existing states
        self.assertRaises(ValidationError, flag.get_clean_state, 100)

        # not int
        self.assertRaises(ValidationError, flag.get_clean_state, 'Not int')

        # None
        self.assertRaises(ValidationError, flag.get_clean_state, None)

    def test_is_flagged(self):
        post = self.post_1
        flag = Flag.objects.get_flag(post)
        flag.count = 0
        flag.save()

        self.assertEqual(False, flag.is_flagged)

        self.set_flag(post)
        flag.refresh_from_db()
        self.assertEqual(True, flag.is_flagged)

    @patch('flag.models.Flag.get_clean_state')
    def test_get_verbose_state(self, mocked_get_clean_state):
        flag = self.flag
        state = flag.State.FLAGGED
        mocked_get_clean_state.return_value = state

        self.assertEqual(flag.get_verbose_state(state), flag.STATE_CHOICES[state-1][1])

        mocked_get_clean_state.return_value = 100
        self.assertIsNone(flag.get_verbose_state(100))

    def test_toggle_state(self):
        post = self.content_object_2
        flag = self.create_flag(post)
        self.assertIsNone(flag.moderator)
        self.assertEqual(flag.state, flag.State.UNFLAGGED.value)

        # toggle states occurs between rejected and resolved only
        self.assertRaises(ValidationError, flag.toggle_state, flag.State.FLAGGED.value, self.moderator)

        flag.toggle_state(flag.State.REJECTED.value, self.moderator)
        self.assertEqual(flag.state, flag.State.REJECTED.value)
        self.assertEqual(flag.moderator, self.moderator)

        # passing RESOLVED state value for the first time
        flag.toggle_state(flag.State.RESOLVED.value, self.moderator)
        self.assertEqual(flag.state, flag.State.RESOLVED.value)

        # passing RESOLVED state value for the second time
        flag.toggle_state(flag.State.RESOLVED.value, self.moderator)
        # state reset to FLAGGED
        self.assertEqual(flag.state, flag.State.FLAGGED.value)

    def test_toggle_flagged_state(self):
        self.flag.increase_count()
        self.flag.toggle_flagged_state()

        self.assertEqual(self.flag.state, Flag.State.FLAGGED.value)

        # set FLAG_ALLOWED to 1
        patch.object(settings, 'FLAG_ALLOWED', 1).start()
        self.flag.count = 0
        self.flag.save()
        self.flag.increase_count()
        self.flag.toggle_flagged_state()

        self.assertEqual(self.flag.state, Flag.State.UNFLAGGED.value)
        # flag once more to toggle the state
        self.flag.increase_count()
        self.flag.toggle_flagged_state()

        self.assertEqual(self.flag.state, Flag.State.FLAGGED.value)


class FlagManagerTest(BaseFlagModelTest):
    def setUp(self):
        super().setUp()
        self.flag = Flag.objects.create(content_object=self.content_object_2, creator=self.post_2.user)

    def test_get_flag(self):
        self.assertEqual(Flag.objects.get_flag(self.post_2), self.flag)

    def test_has_flagged(self):
        user = self.user_2
        post = self.post_2
        self.assertEqual(Flag.objects.has_flagged(user, post), False)

        self.set_flag(self.post_2, user=user)

        self.assertEqual(Flag.objects.has_flagged(user, post), True)

    def test_is_flagged(self):
        post = self.post_1
        flag = Flag.objects.get_flag(post)
        flag.count = 0
        flag.save()

        self.assertEqual(False, Flag.objects.is_flagged(post))

        self.flag.increase_count()
        self.assertEqual(False, Flag.objects.is_flagged(post))


class FlagInstanceModelTest(BaseFlagModelTest):
    def test_create_flag_instance(self):
        self.assertIsNotNone(
            FlagInstance.objects.create(
                flag=self.flag,
                reason=FlagInstance.reason_values[0],
                user=self.user_1
            )
        )


class TestFlagInstanceManager(BaseFlagModelTest):
    def test_clean_when_last_reason_is_used(self):
        flag = self.flag

        with self.assertRaises(ValidationError) as error:
            FlagInstance.objects.create(
                flag=flag,
                reason=FlagInstance.reason_values[-1]
            )
        self.assertEqual(
            error.exception.message_dict['info'],
            ['Please provide some information why you choose to report the content'])

    def test_clean_reason_for_invalid_values(self):
        flag = self.flag
        user = self.user_1
        reason = -1
        info = None

        with self.assertRaises(ValidationError) as error:
            FlagInstance.objects.create_flag(user=user, flag=flag, reason=reason, info=info)

        self.assertEqual(error.exception.messages, [f'{reason} is an invalid reason'])

        reason = 'r'
        with self.assertRaises(ValidationError) as error:
            FlagInstance.objects.create_flag(user=user, flag=flag, reason=reason, info=info)

        self.assertEqual(error.exception.messages, [f'{reason} is an invalid reason'])

        reason = None
        with self.assertRaises(ValidationError) as error:
            FlagInstance.objects.create_flag(user=user, flag=flag, reason=reason, info=info)

        self.assertEqual(error.exception.messages, [f'{reason} is an invalid reason'])

    def test_create_flag_for_flagging_twice(self):
        user = self.user_1
        flag = self.flag
        reason = FlagInstance.reason_values[0]
        info = None

        FlagInstance.objects.create_flag(user, flag, reason, info)
        with self.assertRaises(ValidationError) as error:
            FlagInstance.objects.create_flag(user, flag, reason, info)

        self.assertEqual(error.exception.messages, [f'This content has already been flagged by the user ({user})'])

    def test_delete_flag_without_flagging(self):
        user = self.user_1
        flag = self.flag

        with self.assertRaises(ValidationError) as error:
            FlagInstance.objects.delete_flag(user, flag)

        self.assertEqual(error.exception.messages, [f'This content has not been flagged by the user ({user})'])
